/******************************************************************************/
/*                                                                            */
/* File: code.cpp                                                             */
/*                                                                            */
/* Simple png image file format dump program - dumps out the contents of a    */
/* png file to a text output file (output.txt) that you can then look at.     */
/* Only dumps the chunk sizes and ids, then goes on to dump more extended     */
/* information about the IDAT chunk which contains the actual image           */
/* image information                                                          */
/*                                                                            */
/* Quote: Anyone can code!                                                    */
/*                                                                            */
/******************************************************************************/

// Dump Contents of png image file

#include <string.h>		// memset(..)
#include <stdio.h>		// fopen, fread, sprintf etc
#include <stdarg.h>     // So we can use ... (in dprintf)
#include <stdlib.h>


#include "zlib.h"
#pragma comment(lib, "zdll.lib")


// Saving debug information to a txt file
void dprintf(const char *fmt, ...) 
{
    va_list parms;
    char buf[256];

    // Try to print in the allocated space.
    va_start(parms, fmt);
    vsprintf (buf, fmt, parms);
    va_end(parms);

    // Write the information out to a txt file
    FILE *fp = fopen("output.txt", "a+");
    fprintf(fp, "%s", buf);
    fclose(fp);

}// End dprintf(..)


inline void SwapEndian(unsigned short *data)
{
	*data=((*data&0x00ff)<<8)
		 |((*data&0xff00)>>8);
}

inline void SwapEndian(short *data)
{
	*data=((*data&0x00ff)<<8)
		 |((*data&0xff00)>>8);
}

inline void SwapEndian(unsigned int *data)
{
	*data=((*data&0x000000ff)<<24)
		 |((*data&0x0000ff00)<< 8)
		 |((*data&0x00ff0000)>> 8)
		 |((*data&0xff000000)>>24);
}

inline void SwapEndian(int *data)
{
	*data=((*data&0x000000ff)<<24)
		 |((*data&0x0000ff00)<< 8)
		 |((*data&0x00ff0000)>> 8)
		 |((*data&0xff000000)>>24);
}


static void *png_zalloc(void *opaque, unsigned int items, unsigned int size)
{
    return malloc(items * size);
}

static void png_zfree(void *opaque, void *ptr)
{
    free(ptr);
}


void ReadImageData(FILE *fp, int sizeData)
{
	static z_stream zstream;
	memset(&zstream, 0, sizeof(zstream));

	// init the zlib
	zstream.zalloc = png_zalloc;
	zstream.zfree = png_zfree;
	zstream.opaque = NULL;
	int okay = inflateInit(&zstream);
	dprintf("InflateInit Z_OK(0) : %d\n", okay);


	dprintf("\t[IDAT]Reading Image Data\n");
	dprintf("\t SizeData %d\n", sizeData);

	unsigned char* buf = new unsigned char[sizeData + 100];
	fread(buf, 1, sizeData, fp);

	
	int i = 0;
	dprintf("Compression Flags: 0x%02X\n", buf[i]);	i++;
	dprintf("Additional  Flags: 0x%02X\n", buf[i]);	i++;
	dprintf("Misc Bytes: 0x%02X\n", buf[i]);		i++;
	dprintf("Misc Bytes: 0x%02X\n", buf[i]);		i++;
	dprintf("Misc Bytes: 0x%02X\n", buf[i]);		i++;
	dprintf("Misc Bytes: 0x%02X\n", buf[i]);		i++;
	dprintf("Misc Bytes: 0x%02X\n", buf[i]);		i++;

	for (int y=0; y<8; y++)
	{
		dprintf("Data: ");
		dprintf("0x%02X   - ", buf[i]); i++;

		int lineWidth = 3*8;

		for (int x=0; x<lineWidth; x++)
		{
			dprintf("0x%02X ", buf[i]); i++;
		
		}
		dprintf("\n");
	}


	unsigned int cv = ((unsigned int)buf[i+0] | 
					  ((unsigned int)buf[i+1]<<8) | 
					  ((unsigned int)buf[i+2]<<16) | 
					  ((unsigned int)buf[i+3]<<24));
	SwapEndian(&cv);
	dprintf("Check Value: 0x%08X\n", cv);	i+=4;
	dprintf("i = %d\n\n", i);


	// Decompressed using zlib
	unsigned char outBuf[1024]= {0};

	dprintf("Compressed Data:\n\n");
	unsigned char compressedData[1024] = {0};
	for (int i=0; i<sizeData; i++)
	{
		compressedData[i] = buf[i];

		if (i%(8*3)==0) dprintf("\n");
		dprintf("0x%02X, ", buf[i]);
	}
	dprintf("\n\n");


	int ret = 0;
    zstream.avail_in = sizeData;
    zstream.next_in = compressedData;

	zstream.avail_out = sizeof(outBuf);
    zstream.next_out = (Bytef*)outBuf;


    // decode one line if possible
    while (zstream.avail_in > 0) 
	{
        ret = inflate(&zstream, Z_SYNC_FLUSH); //Z_PARTIAL_FLUSH);
        if (ret != Z_OK && ret != Z_STREAM_END) 
		{
			dprintf("Problem with zlib inflate\n");
            return;
        }
    }

	// Finished!..release zlib
	inflateEnd(&zstream);


	dprintf("-Decompressed Image:\n");
	for (int y=0; y<8; y++)
	{
		int lineWidth = 3*8 + 1;

		dprintf("Data: ");
		dprintf("0x%02X   - ", outBuf[y*lineWidth + 0]);

		for (int x=0; x<lineWidth-1; x++)
		{
			dprintf("0x%02X ", outBuf[y*lineWidth + x + 1]);
		
		}
		dprintf("\n");
	}

	delete[] buf;
}



void ReadPngFile(char* szFileName)
{
	dprintf("Png Dump File: %s\n\n", szFileName);

    FILE *fp;
    fp = fopen(szFileName, "rb");

	// Read 8 byte PNG signature first
	unsigned char sig[8];
	fread(sig, 1, 8, fp);

	// Should be, for a valid PNG file:
	// [0x89, (0x50)'P', (0x4E)'N', (0xD)'G', 0xD, 0xA, 0x1A, 0xA]
	dprintf("First 8 bytes: [0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X, 0x%X]\n",
			sig[0],sig[1],sig[2],sig[3],sig[4],sig[5],sig[6],sig[7]);


	// Next comes the chunks of data which contain the image data
	while (true)
	{
		unsigned int chunkSize;
		unsigned int chunkID;

		
		fread(&chunkSize, 1, sizeof(chunkSize), fp);
		fread(&chunkID,   1, sizeof(chunkID),   fp);

		SwapEndian(&chunkSize); // Get Size of this Chunk
		SwapEndian(&chunkID);	// Get Its Chunk ID Type

		dprintf("[Chunk] ID:0x%X  (Size:% 6d)\n", chunkID, chunkSize);
		bool doneAllChunks = false;

		long chunkPos = ftell(fp);

		switch (chunkID)
		{
			case 0x49444154:	// 'I', 'D', 'A', 'T'
			{
				ReadImageData(fp, chunkSize);
			}
			break;
			case 0x49454e44:	// 'I', 'E', 'N', 'D'
			{
				doneAllChunks=true; // No More Chunks
			}
			break;
			default:
			{
			}
		}

		fseek(fp, chunkPos, SEEK_SET);
		// Jump over this chunk
		fseek(fp, chunkSize, SEEK_CUR); // Skip over its contents

		unsigned int crc;
		fread(&crc,   1, sizeof(crc),   fp); // Read CRC Value
		dprintf("CRC: 0x%08X\n", crc);

		if (doneAllChunks) break;
	}

	// All done extracting information, close file
	fclose(fp);

	dprintf("Finished\n");
}// End ReadPngFile(..)



void main()
{
	ReadPngFile("cross8x8.png");
}// End main()
